import numpy as np


def drectangle(p, x1, x2, y1, y2):
    min = np.minimum
    """Signed distance function for rectangle with corners (x1,y1), (x2,y1),
    (x1,y2), (x2,y2).
    This has an incorrect distance to the four corners but that isn't a big deal
    """
    return -min(min(min(-y1 + p[:, 1], y2 - p[:, 1]), -x1 + p[:, 0]), x2 - p[:, 0])


def dblock(p, x1, x2, y1, y2, z1, z2):
    min = np.minimum
    return -min(
        min(
            min(min(min(-z1 + p[:, 2], z2 - p[:, 2]), -y1 + p[:, 1]), y2 - p[:, 1]),
            -x1 + p[:, 0],
        ),
        x2 - p[:, 0],
    )


def dintersect(d1, d2):
    """Signed distance to set intersection of two regions described by signed
    distance functions d1 and d2.
    Not exact the true signed distance function for the difference,
    for example around corners.
    """
    return np.maximum(d1, d2)


def ddiff(d1, d2):
    """Signed distance to set difference between two regions described by
    signed distance functions d1 and d2.
    Not exact the true signed distance function for the difference,
    for example around corners.
    """
    return np.maximum(d1, -d2)



from mpi4py import MPI
import scipy.sparse as spsparse
from scipy.interpolate import RegularGridInterpolator
import scipy.ndimage
import scipy as sp
import numpy as np
import skfmm

from . import signed_distance_functions as sdf

from .cpp import fast_geometry as gutils

from ..generation.cpp import c_cgal

# cpp implementation of 4x4 determinant calc
dete = gutils.calc_4x4determinant


class SignedDistanceFunctionGenerator:
    """Tool used to build signed distance functions
       from seismic velocity models.
    """

    def __init__(  # noqa: ignore=C901
        self,
        field=None,
        min_threshold=1487,
        max_threshold=99999.0,
        bbox=None,
        gridspacing=None,
        method="union",
        narrow=0.0,
        comm=None,
        flip=False,
    ):
        """Class constructor for :class:`SignedDistanceFunctionGenerator`

        :param field: seismic velocity model on a structured grid
        :type field: array-like with dimensions (nz, nx) or (nz, nx, ny)
        :param min_threshold: seismic velocity min threshold (default=1487 m/s) for which all values > threshold will be meshed
        :type min_threshold: float, optional
        :param max_threshold: seismic velocity max. threshold (default=99999 m/s) for which all values < threshold will be meshed
        :param bbox: bounding box containing domain extents.
        :type bbox: tuple with size (2*dim). For example, in 2D `(zmin, zmax, xmin, xmax)`
        :param gridspacing: space between points of seismic velocity model in meters, required
        :type gridspacing: tuple with size (dim)
        :param method: method used to combine subsection with bounding box. Currently only the default==union.
        :type method: string, optional
        :param narrow: only calculate SDF within a region close to the isocontour default=0.0 i.e., calc'ed everywhere
        :type narrow: float64, optional
        :param comm: MPI4py communicator default==None
        :type comm: MPI4py communicator object, optional
        :param flip: Whether to flip the signed distance function or not (default=False)
        :type flip: boolean

        """
        self.field = field
        self.min_threshold = min_threshold
        self.max_threshold = max_threshold
        self.bbox = bbox
        self.gridspacing = gridspacing
        self.method = method
        self.narrow_band = narrow
        self.flip = flip
        self.SDF = None

        comm = comm or MPI.COMM_WORLD

        if comm.rank == 0:
            print("Building a signed distance function from data...", flush=True)

            if field is None:
                raise ValueError(
                    "field must be specified as a grid of dim containing values to threshold (e.g., MeshSizeFunction object.vp)."
                )

            if bbox is None:
                raise ValueError(
                    "bbox must be a tuple of corner extents in z --> x (--> y) fashion"
                )
            dim = int(len(bbox) / 2)
            if dim < 2 or dim > 3:
                raise ValueError("dimension is invalid, is bbox specified correctly?")
            if gridspacing is None or len(gridspacing) < (dim - 1):
                raise ValueError("gridspacing must be larger than 0.")

            if dim == 2:
                nz, nx = self.field.shape
            elif dim == 3:
                nx, ny, nz = self.field.shape

            phi = np.ones_like(self.field)
            phi[
                np.logical_and(
                    self.field > self.min_threshold, self.field < self.max_threshold
                )
            ] = -1
            # call fast marching method
            d = skfmm.distance(phi, [*self.gridspacing], narrow=self.narrow_band)
            if self.narrow_band > 0:
                d[d > self.narrow_band] = self.narrow_band
                d[d < -self.narrow_band] = -self.narrow_band

            if self.flip:
                print("Flipping the distance sign...", flush=True)
                d *= -1

            # create the grid vectors
            if dim == 2:
                zvec = np.linspace(self.bbox[0], self.bbox[1], nz)
                xvec = np.linspace(self.bbox[2], self.bbox[3], nx)
                z, x = np.meshgrid(zvec, xvec)
                # Apply gaussian filter
                sigma = [10, 10]
                d = sp.ndimage.filters.gaussian_filter(d, sigma, mode="constant")
                interpolant = RegularGridInterpolator(
                    (zvec, xvec), d, bounds_error=False, fill_value=None
                )
            if dim == 3:
                # Apply gaussian filter
                sigma = [10, 10, 10]
                d = sp.ndimage.filters.gaussian_filter(d, sigma, mode="constant")
                zvec = np.linspace(self.bbox[0], self.bbox[1], nz)
                xvec = np.linspace(self.bbox[2], self.bbox[3], nx)
                yvec = np.linspace(self.bbox[4], self.bbox[5], ny)
                x, y, z = np.meshgrid(xvec, yvec, zvec)
                d = d.transpose((2, 0, 1))
                interpolant = RegularGridInterpolator(
                    (zvec, xvec, yvec), d, bounds_error=False, fill_value=None
                )

            # interpolant of level-set
            def SDF1(p):
                return interpolant(p)

            # primary interpolant of box
            if dim == 2:

                def SDF2(p):
                    return sdf.drectangle(p, *self.bbox)

            elif dim == 3:

                def SDF2(p):
                    return sdf.dblock(p, *self.bbox)

            # intersect with box
            if method == "union":

                def SDF(p):
                    d = np.stack((SDF1(p), SDF2(p)), axis=-1)
                    return np.amax(d, 1)

                self.SDF = SDF

            else:
                raise NotImplementedError(
                    "Other methods not yet supported besides union"
                )


def calc_re_ratios(vertices, entities):
    """Calculate radius edge ratios--mesh quality metric

    :param vertices: point coordinates of the mesh vertices'
    :type vertices: numpy.ndarray[float64 x dim]
    :param entities: mesh connectivity table
    :type entities: numpy.ndarray[int x (dim + 1)]

    :return: dihedral angles: array of radius-to-edge ratios
    :rtype: dihedral angles: numpy.ndarray[float64 x 1]
    """
    # circumradius/shortest edge length
    bars = np.concatenate(
        [
            entities[:, [0, 1]],
            entities[:, [1, 2]],
            entities[:, [2, 0]],
            entities[:, [0, 3]],
            entities[:, [1, 3]],
            entities[:, [2, 3]],
        ]
    )
    barvec = vertices[bars[:, 0]] - vertices[bars[:, 1]]
    L = np.sqrt((barvec ** 2).sum(1))
    L = np.reshape(L, (6, len(entities)))
    # min edge length i every tetra
    minL = np.amin(L, axis=0)
    cc = c_cgal.circumballs3(vertices[entities.flatten()])
    r = cc[:, -1]
    return np.sqrt(r) / minL


def remove_external_entities(vertices, entities, extent, dim=2):
    """Remove entities with all dim+1 vertices outside block.

    :param vertices: point coordinates of mesh
    :type vertices: numpy.ndarray[float64 x dim]
    :param entities: mesh connectivity
    :type entities: numpy.ndarray[int x (dim + 1)]
    :param extent: coords. of the local block extents
    :type extent: numpy.ndarray[tuple(float64 x (2*dim))]
    :param dim: dimension of mesh
    :type dim: int, optional

    :return: vertices_new: point coordinates of mesh w/ removed entities
    :rtype: numpy.ndarray[float64 x dim]
    :return: entities_new: mesh connectivity w/ removed entities
    :rtype: numpy.ndarray[int x (dim +1)]
    :return: jx: mapping from old point indexing to new point indexing
    :rtype: numpy.ndarray[int x 1]
    """

    if dim == 2:
        signed_distance = sdf.drectangle(
            vertices[entities.flatten(), :],
            x1=extent[0],
            x2=extent[2],
            y1=extent[1],
            y2=extent[3],
        )
    elif dim == 3:
        signed_distance = sdf.dblock(
            vertices[entities.flatten(), :],
            x1=extent[0],
            x2=extent[3],
            y1=extent[1],
            y2=extent[4],
            z1=extent[2],
            z2=extent[5],
        )
    isOut = np.reshape(signed_distance > 0.0, (-1, (dim + 1)))
    entities_new = entities[(np.sum(isOut, axis=1) != (dim + 1))]
    vertices_new, entities_new, jx = fixmesh(vertices, entities_new, dim=dim)
    return vertices_new, entities_new, jx


def vertex_to_entities(vertices, entities, dim=2):
    """Determine which elements are connected to which vertices.

    :param vertices: point coordinates of mesh vertices
    :type vertices: numpy.ndarray[float64 x dim]
    :param entities: mesh connectivity
    :type entities: numpy.ndarray[int x (dim + 1)]
    :param dim: dimension of mesh
    :type dim: int, optional

    :return: vtoe: indices of entities connected to each vertex
    :rtype: numpy.ndarray[int x 1]

    :return: vtoe_pointer: indices into `vtoe` such that vertex `v` is connected to
                          `vtoe[vtoe_pointer[v]:vtoe_pointer[v+1]]` entities
    :rtype: numpy.ndarray[int x 1]
    """
    num_entities = len(entities)

    ext = np.tile(np.arange(0, num_entities), (dim + 1, 1)).reshape(-1, order="F")
    ve = np.reshape(entities, (-1,))
    ve = np.vstack((ve, ext)).T
    ve = ve[ve[:, 0].argsort(), :]

    idx = np.insert(np.diff(ve[:, 0]), 0, 0)
    vtoe_pointer = np.argwhere(idx)
    vtoe_pointer = np.insert(vtoe_pointer, 0, 0)
    vtoe_pointer = np.append(vtoe_pointer, num_entities * (dim + 1))

    vtoe = ve[:, 1]

    return vtoe, vtoe_pointer


def unique_rows(A, return_index=False, return_inverse=False):
    """Similar to MATLAB's unique(A, 'rows'), this returns B, I, J
    where B is the unique rows of A and I and J satisfy
    A = B[J,:] and B = A[I,:]

    :param  A: array of data
    :type A: numpy.ndarray[int/float64 x N]
    :param return_index: whether to return the indices of unique data
    :type return_index: bool, optional
    :param return_inverse: whether to return the inverse mapping back to A from B.
    :type return_inverse: bool, optional

    :return: B: array of data with duplicates removed
    :rtype: numpy.ndarray[int/float64 x N]
    :return: I: array of indices to unique data B.
    :rtype: numpy.ndarray[int x 1]
    :return: J: array of indices to A from B.
    :rtype: numpy.ndarray[int x 1]
    """
    A = np.require(A, requirements="C")
    assert A.ndim == 2, "array must be 2-dim'l"

    orig_dtype = A.dtype
    ncolumns = A.shape[1]
    dtype = np.dtype((np.character, orig_dtype.itemsize * ncolumns))
    B, I, J = np.unique(A.view(dtype), return_index=True, return_inverse=True)

    B = B.view(orig_dtype).reshape((-1, ncolumns), order="C")

    # There must be a better way to do this:
    if return_index:
        if return_inverse:
            return B, I, J
        else:
            return B, I
    else:
        if return_inverse:
            return B, J
        else:
            return B


def simpvol(p, t):
    """Signed volumes of the simplex elements in the mesh.

    :param p: point coordinates of mesh
    :type p: numpy.ndarray[float64 x dim]
    :param t: mesh connectivity
    :type t: numpy.ndarray[int x (dim + 1)]

    :return: volume: signed volume of entity/simplex.
    :rtype: numpy.ndarray[float64 x 1]
    """

    dim = p.shape[1]
    if dim == 1:
        d01 = p[t[:, 1]] - p[t[:, 0]]
        return d01
    elif dim == 2:
        d01 = p[t[:, 1]] - p[t[:, 0]]
        d02 = p[t[:, 2]] - p[t[:, 0]]
        return (d01[:, 0] * d02[:, 1] - d01[:, 1] * d02[:, 0]) / 2
    elif dim == 3:
        d01 = p[t[:, 1], :] - p[t[:, 0], :]
        d02 = p[t[:, 2], :] - p[t[:, 0], :]
        d03 = p[t[:, 3], :] - p[t[:, 0], :]
        return np.einsum("ij,ij->i", np.cross(d01, d02), d03) / 6
    else:
        raise NotImplementedError


def fixmesh(p, t, ptol=2e-13, dim=2, delunused=False):
    """Remove duplicated/unused vertices and entities and
       ensure orientation of entities is CCW.

    :param p: point coordinates of mesh
    :type p: numpy.ndarray[float64 x dim]
    :param t: mesh connectivity
    :type t: numpy.ndarray[int x (dim + 1)]
    :param ptol: point tolerance to detect duplicates
    :type ptol: float64, optional
    :param dim: dimension of mesh
    :type dim: int, optional
    :param delunused: flag to delete disjoint vertices.
    :type delunused: bool, optional

    :return: p: updated point coordinates of mesh
    :rtype: numpy.ndarray[float64 x dim]
    :return: t: updated mesh connectivity
    :rtype: numpy.ndarray[int x (dim+1)]
    """

    # duplicate vertices
    snap = (p.max(0) - p.min(0)).max() * ptol
    _, ix, jx = unique_rows(np.round(p / snap) * snap, True, True)

    p = p[ix]
    t = jx[t]

    # duplicate entities
    t = np.sort(t, axis=1)
    t = unique_rows(t)

    # delete disjoint vertices
    if delunused:
        pix, _, jx = np.unique(t, return_index=True, return_inverse=True)
        t = np.reshape(jx, (t.shape))
        p = p[pix, :]
        pix = ix[pix]

    # entity orientation is CCW
    flip = simpvol(p, t) < 0
    t[flip, :2] = t[flip, 1::-1]

    return p, t, jx


def simpqual(p, t):
    """Simplex quality radius-to-edge ratio

    :param p: vertex coordinates of mesh
    :type p: numpy.ndarray[float64 x dim]
    :param t: mesh connectivity
    :type t: numpy.ndarray[int x (dim + 1)]

    :return: signed mesh quality: signed mesh quality (1.0 is perfect)
    :rtype: numpy.ndarray[float64 x 1]

    """
    assert p.ndim == 2 and t.ndim == 2 and p.shape[1] + 1 == t.shape[1]

    def length(p1):
        return np.sqrt((p1 ** 2).sum(1))

    a = length(p[t[:, 1]] - p[t[:, 0]])
    b = length(p[t[:, 2]] - p[t[:, 0]])
    c = length(p[t[:, 2]] - p[t[:, 1]])
    r = 0.5 * np.sqrt((b + c - a) * (c + a - b) * (a + b - c) / (a + b + c))
    R = a * b * c / np.sqrt((a + b + c) * (b + c - a) * (c + a - b) * (a + b - c))
    return 2 * r / R


def get_edges(entities, dim=2):
    """Get the undirected edges of mesh in no order (NB: are repeated)

    :param entities: the mesh connectivity
    :type entities: numpy.ndarray[int x (dim+1)]
    :param dim: dimension of the mesh
    :type dim: int, optional

    :return: edges: the edges that make up the mesh
    :rtype: numpy.ndarray[int x 2]
    """

    num_entities = len(entities)
    entities = np.array(entities)
    if dim == 2:
        edges = entities[:, [[0, 1], [0, 2], [1, 2]]]
        edges = edges.reshape((num_entities * 3, 2))
    elif dim == 3:
        edges = entities[:, [[0, 1], [1, 2], [2, 0], [0, 3], [1, 3], [2, 3]]]
        edges = edges.reshape((num_entities * 6, 2))
    return edges


def get_boundary_edges(entities, dim=2):
    """Get the boundary edges of the mesh. Boundary edges only appear (dim-1) times

    :param entities: the mesh connectivity
    :type entities: numpy.ndarray[int x (dim+1)]
    :param dim: dimension of the mesh
    :type dim: int, optional

    :return: boundary_edges: the edges that make up the boundary of the mesh
    :rtype: numpy.ndarray[int x 2]
    """
    edges = get_edges(entities, dim=dim)
    edges = np.sort(edges, axis=1)
    unq, cnt = np.unique(edges, axis=0, return_counts=True)
    boundary_edges = np.array([e for e, c in zip(unq, cnt) if c == (dim - 1)])
    return boundary_edges


def get_winded_boundary_edges(entities):
    """Order the boundary edges of the mesh in a winding fashiono

    :param entities: the mesh connectivity
    :type entities: numpy.ndarray[int x (dim+1)]

    :return: boundary_edges: the edges that make up the boundary of the mesh in a winding order
    :rtype: numpy.ndarray[int x 2]
    """

    boundary_edges = get_boundary_edges(entities)
    _bedges = boundary_edges.copy()

    choice = 0
    isVisited = np.zeros((len(_bedges)))
    ordering = np.array([choice])
    isVisited[choice] = 1

    vStart, vNext = _bedges[choice, :]
    while True:
        locs = np.column_stack(np.where(_bedges == vNext))
        rows = locs[:, 0]
        choices = [row for row in rows if isVisited[row] == 0]
        if len(choices) == 0:
            break
        choice = choices[0]
        ordering = np.append(ordering, [choice])
        isVisited[choice] = 1
        nextEdge = _bedges[choice, :]
        tmp = [v for v in nextEdge if v != vNext]
        vNext = tmp[0]
    boundary_edges = boundary_edges[ordering, :]
    return boundary_edges


def get_boundary_vertices(entities, dim=2):
    """Get the indices of the mesh representing boundary vertices.

    :param entities: the mesh connectivity
    :type entities: numpy.ndarray[int x (dim+1)]
    :param dim: dimension of the mesh
    :type dim: int, optional

    :return: indices: indices into the vertex array that are on the boundary.
    :rtype: numpy.ndarray[float64 x dim]
    """
    if dim == 2:
        b = get_boundary_edges(entities)
    elif dim == 3:
        b = get_boundary_facets(entities)
    indices = np.unique(b.reshape(-1))
    return indices


def get_boundary_entities(vertices, entities, dim=2):
    """Determine the entities that lie on the boundary of the mesh.

    :param vertices: vertex coordinates of mesh
    :type vertices: numpy.ndarray[float64 x dim]
    :param entities: the mesh connectivity
    :type entities: numpy.ndarray[int x (dim+1)]
    :param dim: dimension of the mesh
    :type dim: int, optional

    :return: bele: indices of entities on the boundary of the mesh.
    :rtype: numpy.ndarray[int x 1]
    """
    boundary_vertices = get_boundary_vertices(entities, dim=dim)
    vtoe, ptr = vertex_to_entities(vertices, entities, dim=dim)
    bele = np.array([], dtype=int)
    for vertex in boundary_vertices:
        for ele in zip(vtoe[ptr[vertex] : ptr[vertex + 1]]):
            bele = np.append(bele, ele)
    bele = np.unique(bele)
    return bele


def get_facets(entities):
    """Gets the four facets of each tetrahedral.

    :param entities: the mesh connectivity
    :type entities: numpy.ndarray[int x (dim+1)]

    :return: facets: facets of a tetrahedral entity.
    :rtype: numpy.ndarray[int x 4]
    """
    ix = [[0, 1, 3], [1, 2, 3], [2, 0, 3], [1, 2, 0]]
    return np.array(entities[:, ix]).reshape((len(entities) * 4, 3))


def get_boundary_facets(entities):
    """Get the facets that represent the boundary of a 3D mesh.

    :param entities: the mesh connectivity
    :type entities: numpy.ndarray[int x (dim+1)]

    :return: boundary_facets: facets on the boundary of a 3D mesh.
    :rtype: numpy.ndarray[int x 4]
    """
    facets = get_facets(entities)
    facets = np.sort(facets, axis=1)
    unq, cnt = np.unique(facets, axis=0, return_counts=True)
    boundary_facets = np.array([e for e, c in zip(unq, cnt) if c == 1])
    return boundary_facets


def delete_boundary_entities(vertices, entities, dim=2, minqual=0.10):
    """Delete boundary entities with poor geometric quality (i.e., < min. quality)

    :param vertices: vertex coordinates of mesh
    :type vertices: numpy.ndarray[float64 x dim]
    :param entities: the mesh connectivity
    :type entities: numpy.ndarray[int x (dim+1)]
    :param dim: dimension of the mesh
    :type dim: int, optional
    :param minqual: minimum geometric quality to consider "poor" quality
    :type minqual: float64, optional

    :return: vertices: updated vertex array of mesh
    :rtype: numpy.ndarray[int x dim]
    :return: entities: update mesh connectivity
    :rtype: numpy.ndarray[int x (dim+1)]
    """
    qual = simpqual(vertices, entities)
    bele = get_boundary_entities(vertices, entities, dim=dim)
    qualBou = qual[bele]
    delete = qualBou < minqual
    print(
        "Deleting " + str(np.sum(delete)) + " poor quality boundary entities...",
        flush=True,
    )
    delete = np.argwhere(delete == 1)
    entities = np.delete(entities, bele[delete], axis=0)
    vertices, entities, _ = fixmesh(vertices, entities, delunused=True, dim=dim)
    return vertices, entities


def _sparse(Ix, J, S, shape=None, dtype=None):
    """
    Similar to MATLAB's SPARSE(I, J, S, ...)

    Usage
    -----
    >>> shape = (m, n)
    >>> A = sparse(I, J, S, shape, dtype)
    """

    # Advanced usage: allow J and S to be scalars.
    if np.isscalar(J):
        x = J
        J = np.empty(Ix.shape, dtype=int)
        J.fill(x)
    if np.isscalar(S):
        x = S
        S = np.empty(Ix.shape)
        S.fill(x)

    # Turn these into 1-d arrays for processing.
    S = S.flat
    II = Ix.flat
    J = J.flat
    return spsparse.coo_matrix((S, (II, J)), shape, dtype)


def laplacian2(vertices, entities, max_iter=20, tol=0.01):
    """Move vertices to the average position of their connected neighbors
    with the goal to hopefully improve geometric entity quality.

    :param vertices: vertex coordinates of mesh
    :type vertices: numpy.ndarray[float64 x dim]
    :param entities: the mesh connectivity
    :type entities: numpy.ndarray[int x (dim+1)]
    :param max_iter: maximum number of iterations to perform
    :type max_iter: int, optional
    :param tol: iterations will cease when movement < tol
    :type tol: float64, optional

    :return vertices: updated vertices of mesh
    :rtype: numpy.ndarray[float64 x dim]
    :return: entities: updated mesh connectivity
    :rtype: numpy.ndarray[int x (dim+1)]
    """
    eps = np.finfo(float).eps

    n = len(vertices)

    S = _sparse(
        entities[:, [0, 0, 1, 1, 2, 2]],
        entities[:, [1, 2, 0, 2, 0, 1]],
        1,
        shape=(n, n),
    )
    bnd = get_boundary_vertices(entities)
    edge = get_edges(entities)
    W = np.sum(S, 1)
    if np.any(W == 0):
        print("Invalid mesh. Disjoint vertices found. Returning", flush=True)
        print(np.argwhere(W == 0), flush=True)
        return vertices, entities

    L = np.sqrt(
        np.sum(np.square(vertices[edge[:, 0], :] - vertices[edge[:, 1], :]), axis=1)
    )
    L[L < eps] = eps
    L = L[:, None]
    for it in range(max_iter):
        pnew = np.divide(S * np.matrix(vertices), np.hstack((W, W)))
        pnew[bnd, :] = vertices[bnd, :]
        vertices = pnew
        Lnew = np.sqrt(
            np.sum(np.square(vertices[edge[:, 0], :] - vertices[edge[:, 1], :]), axis=1)
        )
        Lnew[Lnew < eps] = eps
        move = np.amax(np.divide((Lnew - L), Lnew))
        if move < tol:
            print(
                "Movement tolerance reached after " + str(it) + " iterations..exiting",
                flush=True,
            )
            break
        L = Lnew
    vertices = np.array(vertices)
    return vertices, entities


def isManifold(vertices, entities, dim=2):
    """Determine if mesh is manifold by checking for the following:
    1. A boundary edge should be a member of one entity
    2. A non-boundary edge should be a member of two entities
    3. The number of boundary vertices == number of boundary edges

    :param vertices: vertex coordinates of mesh
    :type vertices: numpy.ndarray[float64 x dim]
    :param entities: the mesh connectivity
    :type entities: numpy.ndarray[int x (dim+1)]
    :param dim: dimension of the mesh
    :type dim: int, optional

    :return: isManifold: flag to indicate if the mesh has a manifold boundary.
    :rtype: bool.
    """
    bedges = get_boundary_edges(entities, dim=dim)
    if bedges.size != vertices[np.unique(bedges), :].size:
        print("Mesh has a non-manifold boundary...", flush=True)
        return False
    return True


def vtInEntity2(vertex, entity):
    """
    Does the 2D vertex lie in the entity defined by vertices (x1,y1,x2,y2,x3,y3)?

    :param vertex: vertex coordinates of mesh
    :type vertex: numpy.ndarray[float64 x dim]
    :param entity: connectivity of an entity
    :type entity: numpy.ndarray[int x (dim+1)]

    :return: vtInEntity2: logical flag indicating if it is or isn't.
    :rtype: bool
    """
    (x, y) = vertex
    (x1, y1, x2, y2, x3, y3) = entity
    a = ((y2 - y3) * (x - x3) + (x3 - x2) * (y - y3)) / (
        (y2 - y3) * (x1 - x3) + (x3 - x2) * (y1 - y3)
    )
    b = ((y3 - y1) * (x - x3) + (x1 - x3) * (y - y3)) / (
        (y2 - y3) * (x1 - x3) + (x3 - x2) * (y1 - y3)
    )
    c = 1 - a - b
    # vt lies in entity if and only if 0 <= a <= 1 and 0 <= b <= 1 and 0 <= c <= 1
    return 0 <= a and a <= 1 and 0 <= b and b <= 1 and 0 <= c and c <= 1


def vtInEntity3(vertex, entity):
    """
    Does the 3D vertex lie in the entity defined by vertices (x1,y1,z1,x2,y2,z2,x3,y3,z3)?

    :param vertex: vertex coordinates of mesh
    :type vertex: numpy.ndarray[float64 x dim]
    :param entity: connectivity of an entity
    :type entity: numpy.ndarray[int x (dim+1)]

    :return: vtInEntity3: logical flag indicating if it is or isn't.
    :rtype: bool
    """
    (x, y, z) = vertex
    (x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4) = entity
    signs = np.zeros(5, dtype=int)
    A = np.array(
        [[x1, y1, z1, 1.0], [x2, y2, z2, 1.0], [x3, y3, z3, 1.0], [x4, y4, z4, 1.0]],
    )
    signs[0] = np.sign(dete(A))
    B = np.array(
        [[x, y, z, 1.0], [x2, y2, z2, 1.0], [x3, y3, z3, 1.0], [x4, y4, z4, 1.0]],
    )
    signs[1] = np.sign(dete(B))
    if signs[1] != signs[0]:
        return False
    C = np.array(
        [[x1, y1, z1, 1.0], [x, y, z, 1.0], [x3, y3, z3, 1.0], [x4, y4, z4, 1.0]],
    )
    signs[2] = np.sign(dete(C))
    if signs[2] != signs[1]:
        return False
    D = np.array(
        [[x1, y1, z1, 1.0], [x2, y2, z2, 1.0], [x, y, z, 1.0], [x4, y4, z4, 1.0]],
    )
    signs[3] = np.sign(dete(D))
    if signs[3] != signs[2]:
        return False
    E = np.array(
        [[x1, y1, z1, 1.0], [x2, y2, z2, 1.0], [x3, y3, z3, 1.0], [x, y, z, 1.0]],
    )
    signs[4] = np.sign(dete(E))
    if signs[4] != signs[3]:
        return False
    # point lies in cell if and only if all signs are the same
    return (signs == -1).all() or (signs == 1).all()


def get_centroids(vertices, entities, dim=2):
    """Calculate the centroids of all the entities.

    :param vertex: vertex coordinates of mesh
    :type vertex: numpy.ndarray[float64 x dim]
    :param entities: mesh connectivity
    :type entities: numpy.ndarray[int x (dim+1)]
    :param dim: dimension of mesh
    :type dim: int, optional

    :return: centroids of entities
    :rtype: numpy.ndarray[float64 x dim]
    """
    return vertices[entities].sum(1) / (dim + 1)


def doAnyOverlap(vertices, entities, dim=2):
    """
    Check if any entities connected to boundary of the mesh overlap
    ignoring self-intersections. This routine checks only the 1-ring around
    each entity for potential intersections using barycentric coordinates.

    :param vertex: vertex coordinates of mesh
    :type vertex: numpy.ndarray[float64 x dim]
    :param entities: mesh connectivity
    :type entities: numpy.ndarray[int x (dim+1)]
    :param dim: dimension of mesh
    :type dim: int, optional

    :return: intersections: a list of 2-tuple of entity indices that intersect
    :rtype: List[tuple(num_intersections x 2)]
    """
    vtoe, ptr = vertex_to_entities(vertices, entities, dim=dim)
    # centroids of these elements beles
    cents = get_centroids(vertices, entities, dim=dim)
    # store intersection pairs
    intersections = []
    # for all elements
    for ie, cent in enumerate(cents):
        if ie % 1000 == 0:
            print("INFO: " + str(100.0 * (ie / len(cents))) + " % done...", flush=True)
        # collect all elements neis around element ie
        neis = np.array([], dtype=int)
        for vertex in entities[ie, :]:
            for ele in zip(vtoe[ptr[vertex] : ptr[vertex + 1]]):
                neis = np.append(neis, ele)
        # for all neighboring elements  to element ie
        for iee, ele in enumerate(neis):
            # centroid ie should be in face ie by definition!
            if ie == ele:
                continue
            if dim == 2:
                # does this centroid live inside another neighboring face?
                x1, x2, x3 = vertices[entities[ele, :], 0]
                y1, y2, y3 = vertices[entities[ele, :], 1]
                if vtInEntity2((cent[0], cent[1]), (x1, y1, x2, y2, x3, y3)):
                    # centroid ie is inside face iee
                    print(
                        "Alert: entity "
                        + str(ie)
                        + " intersects with entity "
                        + str(ele)
                        + ". These will be adjusted.",
                        flush=True,
                    )
                    # record all intersection pairs
                    intersections.append((ie, ele))
            elif dim == 3:
                # does this centroid live inside another nei cell?
                x1, x2, x3, x4 = vertices[entities[ele], 0]
                y1, y2, y3, y4 = vertices[entities[ele], 1]
                z1, z2, z3, z4 = vertices[entities[ele], 2]
                if vtInEntity3(
                    (cent[0], cent[1], cent[2]),
                    (x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4),
                ):
                    # centroid ie is inside cell iee
                    print(
                        "Alert: entity "
                        + str(ie)
                        + " intersects with entity "
                        + str(ele)
                        + ". These will be adjusted.",
                        flush=True,
                    )
                    # record all intersection pairs
                    intersections.append((ie, ele))

    return intersections


def linter(vertices, entities, dim=2, minqual=0.10):
    """Remove and check mesh for geometric and toplogical defects.

    :param vertex: vertex coordinates of mesh
    :type vertex: numpy.ndarray[float64 x dim]
    :param entities: mesh connectivity
    :type entities: numpy.ndarray[int x (dim+1)]
    :param dim: dimension of mesh
    :type dim: int, optional
    :param minqual: minimum geometric quality to consider "poor" quality
    :type minqual: float64, optional

    :return vertices: updated mesh vertices
    :rtype: numpy.ndarray[float64 x dim]
    :return: entities: updated mesh connectivity
    :rtype: numpy.ndarray[int x (dim+1)]
    """
    print("Performing mesh linting...", flush=True)
    qual = simpqual(vertices, entities)
    # determine if there's degenerate overlapping elements
    # import time

    # t1 = time.time()
    # intersections = doAnyOverlap(vertices, entities, dim=dim)
    # delete the lower quality in the pair
    # Delete = []
    # For intersect in intersections:
    #    ix = [i for i in intersect]
    #    sel = np.argmin(qual[ix])
    #    delete = np.append(delete, intersect[sel])
    # Delete = np.unique(delete)
    # Print("Deleting " + str(len(delete)) + " overlapped entities", flush=True)
    # Entities = np.delete(entities, delete, axis=0)
    # Print(time.time() - t1)

    # clean up
    vertices, entities, _ = fixmesh(vertices, entities, delunused=True)
    # delete remaining low quality boundary elements
    if dim == 2:
        vertices, entities = delete_boundary_entities(
            vertices, entities, minqual=minqual
        )
        # check for non-manifold boundaries and alert
        _ = isManifold(vertices, entities)
    # calculate final minimum simplex quality
    qual = simpqual(vertices, entities)
    minimum_quality = np.amin(qual)
    print(
        "There are "
        + str(len(vertices))
        + " vertices and "
        + str(len(entities))
        + " elements in the mesh",
        flush=True,
    )
    print("The minimum element quality is " + str(minimum_quality), flush=True)
    return vertices, entities



import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d as a3

from .. import geometry


def plot_tets(points, cells, hold_on=False):
    """
    vizualize tetrahedral mesh
    """
    axes = a3.Axes3D(plt.figure())
    tri = geometry.get_facets(cells)
    vts = points[tri, :]
    tri = a3.art3d.Poly3DCollection(vts)
    tri.set_alpha(0.2)
    tri.set_color("grey")
    axes.add_collection3d(tri)
    axes.plot(points[:, 0], points[:, 1], points[:, 2], "ko")
    axes.set_axis_off()
    if hold_on is not False:
        plt.show()
    return None


def plot_facets(points, facets, color="red", marker="gx", hold_on=False):
    """
    visualize the facets
    """
    axes = a3.Axes3D(plt.figure())
    vts2 = points[facets, :]
    tri2 = a3.art3d.Poly3DCollection(vts2)
    tri2.set_alpha(0.2)
    tri2.set_color(color)
    axes.add_collection3d(tri2)
    axes.plot(points[:, 0], points[:, 1], points[:, 2], marker)
    axes.set_axis_off()
    if hold_on is not False:
        plt.show()
    return None



# encoding: utf-8
# -----------------------------------------------------------------------------
#  Copyright (C) 2020 Keith Jared Roberts

#  Distributed under the terms of the GNU General Public License. You should
#  have received a copy of the license along with this program. If not,
#  see <http://www.gnu.org/licenses/>.

from .sizing import MeshSizeFunction, ReadSegy
from .generation import MeshGenerator
from . import decomp
from . import geometry
from . import migration
from . import plots

__all__ = [
    "geometry",
    "ReadSegy",
    "MeshSizeFunction",
    "MeshGenerator",
    "decomp",
    "migration",
    "plots",
]



#   Copyright (C) 2020 Keith Roberts
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
import time

import numpy as np
from mpi4py import MPI

from . import utils as mutils
from .. import decomp
from .. import migration
from .. import geometry

from .cpp.delaunay_class import DelaunayTriangulation as DT2
from .cpp.delaunay_class3 import DelaunayTriangulation3 as DT3


class MeshGenerator:  # noqa: C901
    """Class constructor for :class:`MeshGenerator`. User can also register their callbacks to
       the sizing function :math:`f(h)` and signed distance function `f(d)` manually.

       :param SizingFunction: A :class:`MeshSizeFunction` object with populated fields and callbacks to `fd` and `fh`.
       :type SizingFunction:  A :class:`MeshSizeFunction` class object. Required if no `fd` or `fh` are passed.
       :param fd: A function that accepts an array of points and returns the signed distance to the boundary of the domain.
       :type fd: A function object, optional if no :class:`SizingFunction` is passed
       :param fh: A call-back function that accepts an array of points and returns an array of desired triangular mesh sizes close by to each point in the passed array.
       :type fh: A function object, optional if no :class:`SizingFunction` is passed.
       :param bbox: bounding box containing domain extents.
       :type bbox: tuple with size (2*dim). For example, in 2D `(zmin, zmax, xmin, xmax)`. Optional if no :class:`SizingFunction` is passed.
       :param hmin: minimum triangular edgelength populating the domain in meters.
       :type hmin: float64,optional if no :class:`SizingFunction` is passed.
       :param pfix: points that you wish you constrain in location.
       :type pfix: nested list [num_fixed_points x dim], optional

       :return: object populated with meta-data.
       :rtype: :class:`MeshGenerator` object
    """

    def __init__(
        self, SizingFunction=None, fd=None, fh=None, bbox=None, hmin=None, pfix=None,
    ):
        self.SizingFunction = SizingFunction
        self.fd = fd
        self.fh = fh
        self.bbox = bbox
        self.hmin = hmin
        self.pfix = pfix

    # SETTERS AND GETTERS
    @property
    def SizingFunction(self):
        return self.__SizingFunction

    @SizingFunction.setter
    def SizingFunction(self, value):
        self.__SizingFunction = value

    @property
    def fd(self):
        return self.__fd

    @fd.setter
    def fd(self, value):
        self.__fd = value

    @property
    def fh(self):
        return self.__fh

    @fh.setter
    def fh(self, value):
        self.__fh = value

    @property
    def bbox(self):
        return self.__bbox

    @bbox.setter
    def bbox(self, value):
        if value is None:
            self.__bbox = value
        else:
            assert (
                len(value) >= 4 and len(value) <= 6
            ), "bbox has wrong number of values. either 4 or 6."
            self.__bbox = value

    @property
    def hmin(self):
        return self.__hmin

    @hmin.setter
    def hmin(self, value):
        self.__hmin = value

    @property
    def pfix(self):
        return self.__pfix

    @pfix.setter
    def pfix(self, value):
        self.__pfix = value

    ### PRIVATE METHODS ###
    def _get_topology(self, dt):
        """ Get points and entities from :clas:`CGAL:DelaunayTriangulation2/3` object"""
        p = dt.get_finite_vertices()
        t = dt.get_finite_cells()
        return p, t

    ### PUBLIC METHODS ###
    def build(  # noqa: ignore=C901
        self,
        SizingFunction=None,
        points=None,
        max_iter=50,
        nscreen=1,
        seed=0,
        comm=None,
        axis=0,
        perform_checks=False,
        mesh_improvement=False,
        min_dh_bound=10,
        max_dh_bound=170,
        enforce_sdf=True,
    ):
        """
         Using callbacks to a sizing function and signed distance field build a simplical mesh.

        :param max_iter: maximum number of iterations (default==50)
        :type max_iter: int, optional
        :param nscreen: output to screen every nscreen iterations (default==1)
        :type nscreen: int, optional
        :param seed: Random seed to ensure results are deterministic (default==0)
        :type seed: int, optional
        :param points: initial point distribution to commence mesh generation (default==None)
        :type points: numpy.ndarray[num_points x dimension], optional
        :param comm: communicator for parallel execution (default==None)
        :type comm: MPI4py communicator object generated when initializing MPI environment, optional.
        :param axis: axis to decomp the domain wrt for parallel execution (default==0)
        :type axis: int, required if parallel.
        :param perform_checks: run serial mesh linting (default==False)
        :type perform_checks: logical, optional
        :param min_dh_bound: minimum dihedral angle allowed (default=5)
        :type min_dh_bound: float64, optional
        :param mesh_improvement: run 3D sliver perturbation mesh improvement (default=False)
        :type mesh_improvement: logical, optional
        :type enforce_sdf: whether to enforce domain boundaries with SDF (default=True)
        :param enforce_sdf: logical, optional

        :return: vertices of simplical mesh
        :rtype: numpy.ndarray[num_points x dimension]
        :return: cells of simplicial mesh
        :rtype: numpy.ndarray[num_cells x (dimension + 1)]
        """
        comm = comm or MPI.COMM_WORLD
        if self.SizingFunction is not None:
            # if :class:`SizingFunction` is passed, grab that data.
            _ef = self.SizingFunction
            fh = _ef.interpolant
            fd = _ef.fd
            bbox = _ef.bbox
            h0 = _ef.hmin
        else:
            # else it had to have been passed to constructor
            fh = self.fh
            fd = self.fd
            bbox = self.bbox
            h0 = self.hmin

        _pfix = self.pfix
        _axis = axis
        _points = points

        # configure parallel computing env.
        if comm.size > 1:
            PARALLEL = True
            rank = comm.Get_rank()
            size = comm.Get_size()
        else:
            PARALLEL = False
            rank = 0
            size = 1

        if mesh_improvement and comm.size > 1:
            PARALLEL = False
            if rank > 0:
                # all ranks exit!
                return True, True

        if mesh_improvement and points is None:
            raise Exception("Mesh improvement requires an initial point set.")

        # set random seed to ensure deterministic results for mesh generator
        if rank == 0:
            print("Setting psuedo-random number seed to " + str(seed), flush=True)
        np.random.seed(seed)

        dim = int(len(bbox) / 2)
        bbox = np.array(bbox).reshape(-1, 2)

        if mesh_improvement and dim == 2:
            raise Exception("Mesh improvement currently on works in 3D")

        L0mult = 1 + 0.4 / 2 ** (dim - 1)
        deltat = 0.1
        geps = 1e-1 * h0
        deps = np.sqrt(np.finfo(np.double).eps) * h0
        min_dh_bound *= math.pi / 180
        max_dh_bound *= math.pi / 180

        # select back-end CGAL call
        if dim == 2:
            DT = DT2
        elif dim == 3:
            DT = DT3

        if _pfix is not None:
            if PARALLEL:
                raise Exception("Fixed points are not yet supported in parallel.")
            else:
                pfix = np.array(_pfix, dtype="d")
                nfix = len(pfix)
            if rank == 0:
                print(
                    "Constraining " + str(nfix) + " fixed points..", flush=True,
                )
        else:
            pfix = np.empty((0, dim))
            nfix = 0

        if _points is None:
            # If the user has NOT supplied points
            if PARALLEL:
                # 1a. Localize mesh size function grid.
                fh = migration.localize_sizing_function(fh, h0, bbox, dim, _axis, comm)
                # 1b. Create initial points in parallel in local box owned by rank
                p = mutils.make_init_points(bbox, rank, size, _axis, h0, dim)
            else:
                # 1. Create initial distribution in bounding box (equilateral triangles)
                p = np.mgrid[
                    tuple(slice(min, max + h0, h0) for min, max in bbox)
                ].astype(float)
                p = p.reshape(dim, -1).T

            # np.savetxt('init.txt', p, delimiter=',')
            # np.savetxt('fd.txt', fd(p), delimiter=',')
            # quit()
            # 2. Remove points outside the region, apply the rejection method
            p = p[fd(p) < geps]  # Keep only d<0 points
            r0 = fh(p)
            r0m = r0.min()
            # Make sure decimation occurs uniformly accross ranks
            if PARALLEL:
                r0m = comm.allreduce(r0m, op=MPI.MIN)
            p = np.vstack(
                (pfix, p[np.random.rand(p.shape[0]) < r0m ** dim / r0 ** dim],)
            )
            if PARALLEL:
                # min x min y min z max x max y max z
                extent = [*np.amin(p, axis=0), *np.amax(p, axis=0)]
                extent[_axis] -= h0
                extent[_axis + dim] += h0
                extents = [comm.bcast(extent, r) for r in range(size)]
        else:
            # If the user has supplied initial points
            if PARALLEL:
                # Domain decompose and localize points
                p = None
                if rank == 0:
                    blocks, extents = decomp.blocker(
                        points=_points, rank=rank, nblocks=size, axis=_axis
                    )
                else:
                    blocks = None
                    extents = None
                # send points to each subdomain
                p, extents = migration.localize_points(blocks, extents, comm, dim)
                extent = extents[rank]
            else:
                p = _points

        N = len(p)

        count = 0
        print(
            "Commencing mesh generation with %d vertices on rank %d." % (N, rank),
            flush=True,
        )

        def dist(p1, p2):
            return np.sqrt(((p1 - p2) ** 2).sum(1))

        pold = None
        while True:

            # 3. (Re)-triangulation by the Delaunay algorithm
            start = time.time()

            # Using CGAL's incremental Delaunay triangulation capabilities.
            if not mesh_improvement:
                dt = DT()
                dt.insert(p.flatten().tolist())
            else:
                if count == 0:
                    dt = DT()
                    dt.insert(p.flatten().tolist())
                else:
                    to_move = np.where(dist(p, pold) > 0)[0]
                    dt.move(to_move.flatten().tolist(), p[to_move].flatten().tolist())

            # Get the current topology of the triangulation
            p, t = self._get_topology(dt)

            pold = p.copy()

            N = p.shape[0]

            if PARALLEL:
                exports = migration.enqueue(extents, p, t, rank, size, dim=dim)
                recv = migration.exchange(comm, rank, size, exports, dim=dim)
                recv_ix = len(recv)
                dt.insert(recv.flatten().tolist())
                p, t = self._get_topology(dt)
                # remove entities will all vertices outside local block
                p, t, inv = geometry.remove_external_entities(p, t, extent, dim=dim,)
                N = p.shape[0]

            # Remove vertices outside the domain
            pmid = p[t].sum(1) / (dim + 1)  # Compute centroids
            t = t[fd(pmid) < -geps]  # Keep interior triangles

            # 4. Describe each bar by a unique pair of nodes
            bars = np.concatenate([t[:, [0, 1]], t[:, [1, 2]], t[:, [2, 0]]])
            if dim == 3:
                bars = np.concatenate(
                    (bars, t[:, [0, 3]], t[:, [1, 3]], t[:, [2, 3]]), axis=0
                )

            bars = mutils.unique_rows(
                np.ascontiguousarray(bars, dtype=np.uint32)
            )  # Bars as node pairs

            # Sliver removal
            if mesh_improvement and count != (max_iter - 1):
                num_move = 0
                # calculate dihedral angles in mesh
                dh_angles = geometry.calc_dihedral_angles(p, t)
                outOfBounds = np.argwhere(
                    (dh_angles[:, 0] < min_dh_bound) | (dh_angles[:, 0] > max_dh_bound)
                )
                eleNums = np.floor(outOfBounds / 6).astype("int")
                eleNums, ix = np.unique(eleNums, return_index=True)

                if count % nscreen == 0:
                    print(
                        "On rank: "
                        + str(rank)
                        + " There are "
                        + str(len(eleNums))
                        + " slivers...",
                        flush=True,
                    )

                move = t[eleNums, 0]
                num_move = move.size
                if num_move == 0:
                    print("Termination reached...No slivers detected!", flush=True)
                    p, t, _ = geometry.fixmesh(p, t, dim=dim, delunused=True)
                    return p, t

                p0, p1, p2, p3 = (
                    p[t[eleNums, 0], :],
                    p[t[eleNums, 1], :],
                    p[t[eleNums, 2], :],
                    p[t[eleNums, 3], :],
                )

                # perturb vector is based on INCREASING circumsphere's radius
                perturb = geometry.calc_circumsphere_grad(p0, p1, p2, p3)
                perturb[np.isinf(perturb)] = 1.0

                # normalize perturbation vector
                perturb_norm = np.sum(np.abs(perturb) ** 2, axis=-1) ** (1.0 / 2)
                perturb /= perturb_norm[:, None]

                # perturb % of local mesh size
                p[move] += 0.10 * h0 * perturb

            if not mesh_improvement:
                # 6. Move mesh points based on bar lengths L and forces F
                barvec = p[bars[:, 0]] - p[bars[:, 1]]  # List of bar vectors
                L = np.sqrt((barvec ** 2).sum(1))  # L = Bar lengths
                L[L == 0] = np.finfo(float).eps
                hbars = fh(p[bars].sum(1) / 2)
                L0 = (
                    hbars
                    * L0mult
                    * ((L ** dim).sum() / (hbars ** dim).sum()) ** (1.0 / dim)
                )
                F = L0 - L
                F[F < 0] = 0  # Bar forces (scalars)
                Fvec = (
                    F[:, None] / L[:, None].dot(np.ones((1, dim))) * barvec
                )  # Bar forces (x,y components)

                Ftot = mutils.dense(
                    bars[:, [0] * dim + [1] * dim],
                    np.repeat([list(range(dim)) * 2], len(F), axis=0),
                    np.hstack((Fvec, -Fvec)),
                    shape=(N, dim),
                )

                Ftot[:nfix] = 0  # Force = 0 at fixed points

                if PARALLEL:
                    if count < max_iter - 1:
                        p += deltat * Ftot
                else:
                    p += deltat * Ftot  # Update node positions

            else:
                # no movement if mesh improvement (from forces)
                maxdp = 0.0

            # 7. Bring outside points back to the boundary
            d = fd(p)
            ix = d > 0  # Find points outside (d>0)

            # if PARALLEL and count is max_iter - 2:
            #   enforce_sdf = False

            if ix.any() and enforce_sdf:

                def deps_vec(i):
                    a = [0] * dim
                    a[i] = deps
                    return a

                dgrads = [(fd(p[ix] + deps_vec(i)) - d[ix]) / deps for i in range(dim)]
                dgrad2 = sum(dgrad ** 2 for dgrad in dgrads)
                dgrad2 = np.where(dgrad2 < deps, deps, dgrad2)
                p[ix] -= (d[ix] * np.vstack(dgrads) / dgrad2).T  # Project

            if count % nscreen == 0 and rank == 0 and not mesh_improvement:
                maxdp = deltat * np.sqrt((Ftot ** 2).sum(1)).max()

            # 8. Number of iterations reached
            if count == max_iter - 1:
                if rank == 0:
                    print(
                        "Termination reached...maximum number of iterations reached.",
                        flush=True,
                    )
                if PARALLEL:
                    p, t = migration.aggregate(p, t, comm, size, rank, dim=dim)
                if rank == 0 and perform_checks:
                    p, t = geometry.linter(p, t, dim=dim)
                break

            # 9. Delete ghost points
            if PARALLEL:
                p = np.delete(p, inv[-recv_ix::], axis=0)
                comm.barrier()

            if count % nscreen == 0 and rank == 0:
                if PARALLEL:
                    print("On rank 0: ", flush=True)
                print(
                    "Iteration #%d, max movement is %f, there are %d vertices and %d cells"
                    % (count + 1, maxdp, len(p), len(t)),
                    flush=True,
                )

            count += 1

            end = time.time()
            if rank == 0 and count % nscreen == 0:
                print("     Elapsed wall-clock time %f : " % (end - start), flush=True)

        return p, t



import numpy as np
import copy
import scipy.sparse as spsparse


def closestNumber(n, m):
    # Find the quotient
    q = int(n / m)

    # 1st possible closest number
    n1 = m * q

    # 2nd possible closest number
    if (n * m) > 0:
        n2 = m * (q + 1)
    else:
        n2 = m * (q - 1)

    # if true, then n1 is the required closest number
    if abs(n - n1) < abs(n - n2):
        return n1

    # else n2 is the required closest number
    return n2


def make_init_points(bbox, rank, size, axis, h0, dim):
    """Create a structured grid in parallel of the entire domain
       Each processor owns a part of the domain.
    """
    _bbox = copy.deepcopy(bbox)

    for i in range(dim):
        if i == axis:
            new_lims = np.linspace(_bbox[i, 0], _bbox[i, 1], size + 1)
            _bbox[i, :] = new_lims[rank : rank + 2]
            if rank != 0:
                # starting point must be lasts + h0
                prev_lims = new_lims[rank - 1 : rank - 1 + 2]
                tmp = np.mgrid[slice(prev_lims[0], prev_lims[1] + h0, h0)]
                _bbox[i, 0] = tmp[-1] + h0

    points = np.mgrid[tuple(slice(min, max + h0, h0) for min, max in _bbox)].astype(
        float
    )
    points = points.reshape(dim, -1).T
    return points


def __setdiff_rows(A, B, return_index=False):
    """
    Similar to MATLAB's setdiff(A, B, 'rows'), this returns C, I
    where C are the row of A that are not in B and I satisfies
    C = A[I,:].

    Returns I if return_index is True.
    """
    A = np.require(A, requirements="C")
    B = np.require(B, requirements="C")

    assert A.ndim == 2, "array must be 2-dim'l"
    assert B.ndim == 2, "array must be 2-dim'l"
    assert A.shape[1] == B.shape[1], "arrays must have the same number of columns"
    assert A.dtype == B.dtype, "arrays must have the same data type"

    # NumPy provides setdiff1d, which operates only on one dimensional
    # arrays. To make the array one-dimensional, we interpret each row
    # as being a string of characters of the appropriate length.
    orig_dtype = A.dtype
    ncolumns = A.shape[1]
    dtype = np.dtype((np.character, orig_dtype.itemsize * ncolumns))
    C = (
        np.setdiff1d(A.view(dtype), B.view(dtype))
        .view(A.dtype)
        .reshape((-1, ncolumns), order="C")
    )
    if return_index:
        raise NotImplementedError
    else:
        return C


def unique_rows(ar):
    ar_row_view = ar.view("|S%d" % (ar.itemsize * ar.shape[1]))
    _, unique_row_indices = np.unique(ar_row_view, return_index=True)
    ar_out = ar[unique_row_indices]
    return ar_out


def dense(Ix, J, S, shape=None, dtype=None):
    """
    Similar to MATLAB's SPARSE(I, J, S, ...), but instead returning a
    dense array.

    Usage
    -----
    >>> shape = (m, n)
    >>> A = dense(I, J, S, shape, dtype)
    """

    # Advanced usage: allow J and S to be scalars.
    if np.isscalar(J):
        x = J
        J = np.empty(Ix.shape, dtype=int)
        J.fill(x)
    if np.isscalar(S):
        x = S
        S = np.empty(Ix.shape)
        S.fill(x)

    # Turn these into 1-d arrays for processing.
    S = S.flat
    II = Ix.flat
    J = J.flat
    return spsparse.coo_matrix((S, (II, J)), shape, dtype).toarray()



#   Copyright (C) 2020 Keith Roberts
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import warnings

from mpi4py import MPI
import numpy as np
from scipy.interpolate import RegularGridInterpolator
from scipy import ndimage
import h5py
import matplotlib.pyplot as plt

import segyio

from ..geometry import signed_distance_functions as sdf
from .cpp import limgrad


def ReadSegy(fname):
    """Read a velocity model from a SEG-y file"""
    with segyio.open(fname, ignore_geometry=True) as f:
        nz, nx = len(f.samples), len(f.trace)
        vp = np.zeros(shape=(nz, nx))
        for index, trace in enumerate(f.trace):
            vp[:, index] = trace
        return np.flipud(vp)


class MeshSizeFunction:
    """The :class:`MeshSizeFunction` is used to build a rectangular or cubic isotropic mesh size function :math:`f(h)`.
       and assumes the domain is a rectangle (2D) or cube (3D).
    """

    def __init__(
        self,
        bbox,
        hmin,
        velocity_grid,
        units="m-s",
        wl=0.0,
        freq=5.0,
        grad=0.0,
        space_order=1,
        hmax=np.inf,
        dt=0.0,
        cr_max=1.0,
        grade=0.0,
        nx=None,
        ny=None,
        nz=None,
        domain_ext=0.0,
        padstyle="edge",
    ):
        """Class constructor for :class:`MeshSizeFunction`

        :param bbox: bounding box containing domain extents.
        :type bbox: tuple with size (2*dim). For example, in 2D `(zmin, zmax, xmin, xmax)`
        :param hmin: minimum triangular edgelength populating the domain in meters.
        :type hmin: float64
        :param velocity_grid: a grid of velocity values.
        :type velocity_grid: array-like, numpy.ndarray()
        :param units: units of the velocity model (either `m-s` or `km-s`)
        :type units: str, optional, default=`m-s`
        :param wl: number of vertices per wavelength for a given :math:`f_{max}`
        :type wl: int, optional
        :param grad: the resolution in m nearby sharp gradients in velociy.
        :type grad: float64, optional
        :param freq: :math:`f_{max}` in hertz for which to estimate `wl`
        :type freq: float64, optional
        :param space_order: the polynomial order of the basis functions.
        :type space_order: int, optional
        :param hmax: maximum mesh size in meters allowed in the domain
        :type hmax: float64, optional
        :param dt: theoretical maximum stable timestep in seconds given Courant number `Cr`
        :type dt: float64, optional
        :param cr_max: `dt` is stable with this Courant number.
        :type cr_max: float64, optional
        :param grade: maximum allowable variation in mesh size in decimal percent.
        :type grade: float64, optional
        :param domain_ext: width of domain extension in `-z`, `+x`, `-x`, `+y`, `-y` directions
        :type domain_ext: float64, optional
        :param padstyle: method to pad velocity in the domain extension
        :type padstyle: str, optional, `edge`, `linear_ramp`, `constant`

        :return: object populated with meta-data, :math:`f(h)`, and a :math:`f(d)`.
        :rtype: :class:`MeshSizeFunction` object
        """

        self.bbox = bbox
        self.velocity_grid = velocity_grid
        self.dim = int(len(self.bbox) / 2)
        self.width = bbox[3] - bbox[2]
        self.depth = bbox[1] - bbox[0]
        if self.dim == 3:
            self.length = bbox[5] - bbox[4]
        self.spacingZ = None
        self.spacingX = None
        self.units = units
        self.hmin = hmin
        self.hmax = hmax
        self.wl = wl
        self.freq = freq
        self.grad = grad
        self.space_order = space_order
        self.dt = dt
        self.cr_max = cr_max
        self.grade = grade
        self.fh = None
        self.fd = None
        self.nx = nx
        self.ny = ny
        self.nz = nz
        self.domain_ext = domain_ext
        self.padstyle = padstyle
        self.interpolant = None

    ### SETTERS AND GETTERS ###

    @property
    def interpolant(self):
        return self.__interpolant

    @interpolant.setter
    def interpolant(self, value):
        self.__interpolant = value

    @property
    def fh(self):
        return self.__fh

    @fh.setter
    def fh(self, value):
        self.__fh = value

    @property
    def fd(self):
        return self.__fd

    @fd.setter
    def fd(self, value):
        self.__fd = value

    @property
    def bbox(self):
        return self.__bbox

    @bbox.setter
    def bbox(self, value):
        assert (
            len(value) >= 4 and len(value) <= 6
        ), "bbox has wrong number of values. either 4 or 6."
        self.__bbox = value

    @property
    def hmin(self):
        return self.__hmin

    @hmin.setter
    def hmin(self, value):
        assert value > 0.0, "hmin must be non-zero"
        self.__hmin = value

    @property
    def dim(self):
        return self.__dim

    @dim.setter
    def dim(self, value):
        assert value == 2 or value == 3, "dim must be either 2 or 3"
        self.__dim = value

    @property
    def velocity_grid(self):
        return self.__velocity_grid

    @velocity_grid.setter
    def velocity_grid(self, value):
        if value is None:
            raise ValueError("Velocity grid must be a numpy array")
        if np.amin(value) < 1000:
            warnings.warn("Min. velocity < 1000 m-s. Units may be incorrect.")
        if np.amax(value) > 10000:
            warnings.warn("Max. velocity > 10,000 m-s. Units may be incorrect.")
        self.__velocity_grid = value

    @property
    def units(self):
        return self.__units

    @units.setter
    def units(self, value):
        assert value == "m-s" or value == "km-s", "units are not compatible"
        self.__units = value

    @property
    def wl(self):
        return self.__wl

    @wl.setter
    def wl(self, value):
        self.__wl = value

    @property
    def grad(self):
        return self.__grad

    @grad.setter
    def grad(self, value):
        self.__grad = value

    @property
    def freq(self):
        return self.__freq

    @freq.setter
    def freq(self, value):
        self.__freq = value

    @property
    def space_order(self):
        return self.__space_order

    @space_order.setter
    def space_order(self, value):
        self.__space_order = value

    @property
    def hmax(self):
        return self.__hmax

    @hmax.setter
    def hmax(self, value):
        self.__hmax = value

    @property
    def dt(self):
        return self.__dt

    @dt.setter
    def dt(self, value):
        assert value >= 0, "dt must be > 0"
        self.__dt = value

    @property
    def cr_max(self):
        return self.__cr_max

    @cr_max.setter
    def cr_max(self, value):
        assert value >= 0, "Cr_max must be > 0"
        self.__cr_max = value

    @property
    def grade(self):
        return self.__grade

    @grade.setter
    def grade(self, value):
        assert value >= 0, "grade must be > 0"
        self.__grade = value

    @property
    def domain_ext(self):
        return self.__domain_ext

    @domain_ext.setter
    def domain_ext(self, value):
        assert value >= 0, "domain extent must be > 0"
        self.__domain_ext = value

    @property
    def padstyle(self):
        return self.__padstyle

    @padstyle.setter
    def padstyle(self, value):
        assert value == "edge" or value == "constant" or value == "linear_ramp"
        self.__padstyle = value

    # ---PUBLIC METHODS---#

    def build(self, comm=None):  # noqa: C901

        """Builds the isotropic mesh size function according
            to the user arguments that were passed.
        """
        comm = comm or MPI.COMM_WORLD
        if comm is not None:
            rank = comm.Get_rank()
            size = comm.Get_size()
        else:
            rank = 0
            size = 1

        if rank == 0:
            self.__ReadVelocityModel()
            _vp = self.velocity_grid

            _bbox = self.bbox
            _dim = self.dim
            _width = self.width
            _nz = self.nz
            _nx = self.nx
            if _dim == 3:
                _ny = self.ny
            _domain_ext = self.domain_ext

            _hmax = self.hmax
            _hmin = self.hmin
            _grade = self.grade

            _wl = self.wl
            _freq = self.freq
            _space_order = self.space_order

            _grad = self.grad

            _dt = self.dt
            _cr_max = self.cr_max

            if _dim == 2:
                hh_m = np.zeros(shape=(_nz, _nx), dtype=np.float32) + _hmin
            if _dim == 3:
                hh_m = np.zeros(shape=(_nz, _nx, _ny), dtype=np.float32) + _hmin

            if _wl > 0:
                if rank == 0:
                    print(
                        "Mesh sizes will be built to resolve an estimate of wavelength with "
                        + str(_wl)
                        + " vertices...",
                        flush=True,
                    )
                hh_m = _vp / (_freq * _wl)

            if _grad > 0:
                if rank == 0:
                    print(
                        "Refining mesh sizes near sharp velocity gradients...",
                        flush=True,
                    )
                if _dim == 2:
                    win_rows, win_cols = 100, 100
                    win_mean = ndimage.uniform_filter(_vp, (win_rows, win_cols))
                    win_sqr_mean = ndimage.uniform_filter(
                        _vp ** 2, (win_rows, win_cols)
                    )
                elif _dim == 3:
                    win_rows, win_cols, win_cols2 = 100, 100, 100
                    win_mean = ndimage.uniform_filter(
                        _vp, (win_rows, win_cols, win_cols2)
                    )
                    win_sqr_mean = ndimage.uniform_filter(
                        _vp ** 2, (win_rows, win_cols, win_cols2)
                    )
                win_var = win_sqr_mean - win_mean ** 2
                # normalize variance to [0,1]
                win_var /= np.amax(win_var)
                win_var -= np.amin(win_var)
                tmp = _grad / (win_var + 0.10)
                hh_m = np.minimum(tmp, hh_m)

            # enforce min (and optionally max) sizes
            hh_m = np.where(hh_m < _hmin, _hmin, hh_m)
            if _hmax < np.inf:
                if rank == 0:
                    print("Enforcing maximum mesh resolution...", flush=True)
                hh_m = np.where(hh_m > _hmax, _hmax, hh_m)

            # adjust mesh res. based on the CFL limit so cr < cr_max
            # NB: considering the space order p
            if _dt > 0:
                if rank == 0:
                    print(
                        "Enforcing timestep of " + str(_dt) + " seconds...", flush=True
                    )
                # assume isotropic mesh resolution
                cr_old = (_vp * _dt) / (_dim * hh_m)
                # divide maximum Courant by desired space order
                _cr_max = _cr_max / (_dim * _space_order)
                # determine resolution that statisfies desired _cr_max
                dxn = (_vp * _dt) / (_dim * _cr_max)
                # edit mesh size function to avoid violations of cr
                hh_m = np.where(cr_old > _cr_max, dxn, hh_m)

            # grade the mesh sizes
            if _grade > 0:
                if rank == 0:
                    # TODO: grader needs to consider varying constant resolution by dimension
                    print("Enforcing mesh gradation in sizing function...", flush=True)
                    hh_m = self.hj(hh_m, _width / _nx, 10000)

            # Domain extension
            # edit the bbox to reflect the new domain size
            if _domain_ext > 0:
                self = self.__CreateDomainExtension()  # edit the bbox
                hh_m = self.__EditMeshSizeFunction(
                    hh_m, rank
                )  # edit the sizing function

            # construct a interpolator object to be queried during mesh generation
            if rank == 0:
                print("Building a gridded interpolant...", flush=True)

            if _dim == 2:
                z_vec, x_vec = self.__CreateDomainVectors()

            if _dim == 3:
                z_vec, x_vec, y_vec = self.__CreateDomainVectors()

            assert np.all(hh_m > 0.0), "Mesh size_function must be strictly positive."

            if _dim == 2:
                self.interpolant = RegularGridInterpolator(
                    (z_vec, x_vec), hh_m, bounds_error=False, fill_value=None
                )
            if _dim == 3:
                self.interpolant = RegularGridInterpolator(
                    (z_vec, x_vec, y_vec), hh_m, bounds_error=False, fill_value=None
                )
            # Python can't bcast pickles so this is done after in parallel
            if size == 1:
                # create a mesh size function interpolant
                self.fh = lambda p: self.interpolant(p)

                _bbox = self.bbox

                # signed distance function for a rectangle
                def fdd(p):
                    return sdf.drectangle(p, _bbox[0], _bbox[1], _bbox[2], _bbox[3])

                # signed distance function for a cube
                def fdd2(p):
                    return sdf.dblock(
                        p, _bbox[0], _bbox[1], _bbox[2], _bbox[3], _bbox[4], _bbox[5]
                    )

                if _dim == 2:
                    self.fd = lambda p: fdd(p)
                if _dim == 3:
                    self.fd = lambda p: fdd2(p)

        # if parallel
        if size > 1:
            self._construct_lambdas(comm)
        return self

    def _construct_lambdas(self, comm):
        """
        Build lambda fields (for parallel only) for
        SDF and mesh size function
        """
        self.bbox = comm.bcast(self.bbox, 0)

        _dim = self.dim
        _bbox = self.bbox

        # create a mesh size function interpolant
        self.fh = lambda p: self.interpolant(p)

        def fdd(p):
            return sdf.drectangle(p, _bbox[0], _bbox[1], _bbox[2], _bbox[3])

        def fdd2(p):
            return sdf.dblock(
                p, _bbox[0], _bbox[1], _bbox[2], _bbox[3], _bbox[4], _bbox[5]
            )

        # create a signed distance function
        if _dim == 2:
            self.fd = lambda p: fdd(p)
        if _dim == 3:
            self.fd = lambda p: fdd2(p)
        return self

    def plot(self, stride=5, comm=None):
        """ Plot the isotropic mesh size function"""
        comm = comm or MPI.COMM_WORLD
        if comm.rank == 0:
            _dim = self.dim
            _fh = self.fh
            _domain_ext = self.domain_ext
            _width = self.width
            _depth = self.depth

            if _dim == 2:
                zg, xg = self.__CreateDomainMatrices()
                sz1z, sz1x = zg.shape
                _zg = np.reshape(zg, (-1, 1))
                _xg = np.reshape(xg, (-1, 1))
                hh = _fh((_zg, _xg))
                hh = np.reshape(hh, (sz1z, sz1x))

                fig, ax = plt.subplots()
                plt.pcolormesh(
                    xg[0::stride], zg[0::stride], hh[0::stride], edgecolors="none"
                )
                if _domain_ext > 0:
                    rect = plt.Rectangle(
                        (0, -_depth + _domain_ext),
                        _width - 2 * _domain_ext,
                        _depth - _domain_ext,
                        fill=False,
                        edgecolor="black",
                    )
                    ax.add_patch(rect)

                plt.title("Isotropic mesh sizes")
                plt.colorbar(label="mesh size (m)")
                plt.xlabel("x-direction (m)")
                plt.ylabel("z-direction (m)")
                ax.axis("equal")
                # ax.set_xlim(0 - _domain_ext, _width)
                # ax.set_ylim(-_depth, 0)
                plt.show()
            elif _dim == 3:
                print("visualization in 3D not yet supported!")
        return None

    def GetDomainMatrices(self):
        """ Accessor to private method"""
        zg, xg = self.__CreateDomainMatrices()
        return zg, xg

    def hj(self, fun, elen, imax):
        """Call-back to the cpp gradient limiter code """
        _dim = self.dim
        _nz = self.nz
        _nx = self.nx
        if _dim == 3:
            _ny = self.ny
        _grade = self.grade

        ffun = fun.flatten("F")
        ffun_list = ffun.tolist()
        if _dim == 2:
            tmp = limgrad([_nz, _nx, 1], elen, _grade, imax, ffun_list)
        if _dim == 3:
            tmp = limgrad([_nz, _nx, _ny], elen, _grade, imax, ffun_list)
        tmp = np.asarray(tmp)
        if _dim == 2:
            fun_s = np.reshape(tmp, (_nz, _nx), "F")
        if _dim == 3:
            fun_s = np.reshape(tmp, (_nz, _nx, _ny), "F")
        return fun_s

    def WriteVelocityModel(self, ofname, comm=None):
        """Writes a velocity model as a hdf5 file for use in Spyro """
        comm = comm or MPI.COMM_WORLD
        if comm.rank == 0:
            _dim = self.dim
            _vp = self.velocity_grid
            _nz = self.nz
            _nx = self.nx
            if _dim == 3:
                _ny = self.ny
            _domain_ext = self.domain_ext
            _spacingZ = self.spacingZ
            _spacingX = self.spacingX
            _padstyle = self.padstyle
            nnz = int(_domain_ext / _spacingZ)
            nnx = int(_domain_ext / _spacingX)
            # create domain extension in velocity model
            mx = np.amax(_vp)
            if _dim == 2:
                if _padstyle == "edge":
                    _vp = np.pad(_vp, ((nnz, 0), (nnx, nnx)), "edge")
                elif _padstyle == "constant":
                    # set to maximum value in domain
                    _vp = np.pad(
                        _vp,
                        ((nnz, 0), (nnx, nnx)),
                        "constant",
                        constant_values=(mx, mx),
                    )
                elif _padstyle == "linear_ramp":
                    # linearly ramp to maximum value in domain
                    _vp = np.pad(
                        _vp, ((nnz, 0), (nnx, nnx)), "linear_ramp", end_values=(mx, mx)
                    )
            if _dim == 3:
                if _padstyle == "edge":
                    _vp = np.pad(_vp, ((nnz, 0), (nnx, nnx), (nnx, nnx)), "edge")
                elif _padstyle == "linear_ramp":
                    _vp = np.pad(
                        _vp,
                        ((nnz, 0), (nnx, nnx), (nnx, nnx)),
                        "linear_ramp",
                        end_values=(mx, mx),
                    )
                elif _padstyle == "constant":
                    # set to maximum value in domain
                    _vp = np.pad(
                        _vp,
                        ((nnz, 0), (nnx, nnx), (nnx, nnx)),
                        "constant",
                        constant_values=(mx, mx),
                    )

            print(_vp.shape)
            _nz += nnz  # only bottom
            _nx += nnx * 2  # left and right
            if _dim == 3:
                _ny += nnx * 2  # behind and in front

            ofname += ".hdf5"
            print("Writing velocity model " + ofname, flush=True)
            with h5py.File(ofname, "w") as f:
                f.create_dataset("velocity_model", data=_vp, dtype="f")
                if _dim == 2:
                    f.attrs["shape"] = (_nz, _nx)
                elif _dim == 3:
                    f.attrs["shape"] = (_nz, _nx, _ny)
                f.attrs["units"] = "m/s"

    def SaveMeshSizeFunctionOptions(self, ofname):
        " Save your mesh size function options as a hdf5 file" ""
        ofname += ".hdf5"
        print("Saving mesh size function options " + ofname)
        with h5py.File(ofname, "a") as f:
            f.attrs["bbox"] = self.bbox
            f.attrs["units"] = self.units
            f.attrs["hmin"] = self.hmin
            f.attrs["hmax"] = self.hmax
            f.attrs["wl"] = self.wl
            f.attrs["freq"] = self.freq
            f.attrs["dt"] = self.dt
            f.attrs["cr_max"] = self.cr_max
            f.attrs["grade"] = self.grade
            f.attrs["nx"] = self.nx
            if self.dim == 3:
                f.attrs["ny"] = self.ny
            f.attrs["nz"] = self.nz
            f.attrs["domain_ext"] = self.domain_ext
            f.attrs["padstyle"] = self.padstyle

    # ---PRIVATE METHODS---#

    def __ReadVelocityModel(self):
        """Parses a numpy.ndarray velocity model"""
        # nz, nx, ny and spacingZ, and spacingX
        sd = self.dim
        if sd == 2:
            self.nz, self.nx = self.velocity_grid.shape
        elif sd == 3:
            self.nz, self.nx, self.ny = self.velocity_grid.shape
        self.spacingZ = self.depth / self.nz
        self.spacingX = self.width / self.nx

        if self.__units == "km-s":
            print("Converting model velocities to m-s...")
            self.velocity_grid *= 1e3

    def __CreateDomainVectors(self):
        """ it is what it is """
        _dim = self.dim
        _nz = self.nz
        _nx = self.nx
        if _dim == 3:
            _ny = self.ny
        _domain_ext = self.domain_ext
        _spacingZ = self.spacingZ
        _spacingX = self.spacingX
        _bbox = self.bbox

        # if domain extension is enabled, we augment the domain vectors
        nnz = int(_domain_ext / _spacingZ)
        nnx = int(_domain_ext / _spacingX)
        if _domain_ext > 0:
            _nz += nnz  # only bottom
            _nx += nnx * 2  # left and right
            if _dim == 3:
                _ny += nnx * 2  # behind and in front

        zvec = np.linspace(_bbox[0], _bbox[1], _nz, dtype=np.float32)
        xvec = np.linspace(
            _bbox[2] - _domain_ext, _bbox[3] + _domain_ext, _nx, dtype=np.float32
        )

        if _dim == 2:
            return zvec, xvec
        elif _dim == 3:
            yvec = np.linspace(
                _bbox[4] - _domain_ext, _bbox[5] + _domain_ext, _ny, dtype=np.float32
            )
            return zvec, xvec, yvec

    def __CreateDomainMatrices(self):
        _dim = self.dim
        if _dim == 2:
            zvec, xvec = self.__CreateDomainVectors()
            zg, xg = np.meshgrid(zvec, xvec, indexing="ij")
            return zg, xg
        elif _dim == 3:
            zvec, xvec, yvec = self.__CreateDomainVectors()
            xg, yg, zg = np.meshgrid(
                xvec, yvec, zvec, indexing="ij", sparse=True, copy=False
            )
            return zg, xg, yg

    def __EditMeshSizeFunction(self, hh_m, rank):
        """ Edits sizing function to support domain extension of variable width """
        _domain_ext = self.domain_ext
        _dim = self.dim
        _hmax = self.hmax
        _spacingZ = self.spacingZ
        _spacingX = self.spacingX
        _padstyle = self.padstyle

        nnz = int(_domain_ext / _spacingZ)
        nnx = int(_domain_ext / _spacingX)

        if rank == 0:
            print(
                "Including a " + str(_domain_ext) + " meter domain extension...",
                flush=True,
            )
            print(
                "Using the padstyle " + _padstyle + " to extend mesh resolution.",
                flush=True,
            )
        if _dim == 2:
            mx = np.amax(hh_m)
            if _padstyle == "edge":
                hh_m = np.pad(hh_m, ((nnz, 0), (nnx, nnx)), "edge")
            elif _padstyle == "constant":
                # set to maximum value in domain
                hh_m = np.pad(
                    hh_m, ((nnz, 0), (nnx, nnx)), "constant", constant_values=(mx, mx)
                )
            elif _padstyle == "linear_ramp":
                # linearly ramp to maximum value in domain
                hh_m = np.pad(
                    hh_m, ((nnz, 0), (nnx, nnx)), "linear_ramp", end_values=(mx, mx)
                )
            hh_m = np.where(hh_m > _hmax, _hmax, hh_m)
            return hh_m
        if _dim == 3:
            mx = np.amax(hh_m)
            if _padstyle == "edge":
                hh_m = np.pad(hh_m, ((nnz, 0), (nnx, nnx), (nnx, nnx)), "edge")
            elif _padstyle == "linear_ramp":
                # linearly ramp to maximum value in domain
                hh_m = np.pad(
                    hh_m,
                    ((nnz, 0), (nnx, nnx), (nnx, nnx)),
                    "linear_ramp",
                    end_values=(mx, mx),
                )
            else:
                raise ValueError("3D pad style currently not supported yet")
            hh_m = np.where(hh_m > _hmax, _hmax, hh_m)
            return hh_m

    def __CreateDomainExtension(self):
        """Edit bbox to reflect domain extension"""
        _dim = self.dim
        _bbox = self.bbox
        _domain_ext = self.domain_ext
        if _domain_ext > 0:
            if _dim == 2:
                bbox_new = (
                    _bbox[0] - _domain_ext,
                    _bbox[1],
                    _bbox[2] - _domain_ext,
                    _bbox[3] + _domain_ext,
                )
            if _dim == 3:
                bbox_new = (
                    _bbox[0] - _domain_ext,
                    _bbox[1],
                    _bbox[2] - _domain_ext,
                    _bbox[3] + _domain_ext,
                    _bbox[4] - _domain_ext,
                    _bbox[5] + _domain_ext,
                )

            self.bbox = bbox_new
            self.width = bbox_new[3] - bbox_new[2]
            self.depth = bbox_new[1] - bbox_new[0]
            if _dim == 3:
                self.length = bbox_new[5] - bbox_new[4]
        return self



import numpy as np
import zipfile

import meshio
from mpi4py import MPI

import SeismicMesh

comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.Get_rank()

# Serial or parallel 3d mesh generation building a mesh roughly 1.6 million cells.
# Warning: In serial, this example takes roughly 20 minutes...


def example_3D():

    # The velocity model was downloaded from here:
    # https://s3.amazonaws.com/open.source.geoscience/open_data/seg_eage_models_cd/Salt_Model_3D.tar.gz

    # Dimensions of model
    nx, ny, nz = 676, 676, 210

    path = "velocity_models/Salt_Model_3D/3-D_Salt_Model/VEL_GRIDS/"
    # Extract Saltf@@ from SALTF.ZIP
    zipfile.ZipFile(path + "SALTF.ZIP", "r").extract("Saltf@@", path=path)

    # Load data
    with open(path + "Saltf@@", "r") as file:
        vp = np.fromfile(file, dtype=np.dtype("float32").newbyteorder(">"))
        vp = vp.reshape(nx, ny, nz, order="F")
        vp = np.flipud(vp.transpose((2, 0, 1)))  # z, x and then y

    # Bounding box describing domain extents (corner coordinates)
    bbox = (-4200, 0, 0, 13520, 0, 13520)

    # Construct mesh sizing object from velocity model
    ef = SeismicMesh.MeshSizeFunction(
        bbox=bbox,
        velocity_grid=vp,
        dt=0.001,
        freq=2,
        wl=5,
        grade=0.25,
        hmin=150,
        hmax=5e3,
        domain_ext=250,
        padstyle="linear_ramp",
    )

    # Build mesh size function (in parallel)
    ef = ef.build()

    # Write to disk for later use
    ef.WriteVelocityModel("EAGE_Salt")

    # Construct a mesh generator object
    mshgen = SeismicMesh.MeshGenerator(ef)

    # Build the mesh
    points, cells = mshgen.build(max_iter=75, axis=1)

    # Do mesh improvement in serial to bound lower dihedral angle to >= 5 degrees
    points, cells = mshgen.build(
        points=points, mesh_improvement=True, max_iter=50, min_dh_bound=5,
    )

    if rank == 0:
        # Write to disk (see meshio package for more details)
        meshio.write_points_cells(
            "EAGE_Salt.vtk", points / 1000.0, [("tetra", cells)],
        )

        # Write to gmsh22 format (quite slow)
        meshio.write_points_cells(
            "EAGE_Salt.msh",
            points / 1000,
            [("tetra", cells)],
            file_format="gmsh22",
            binary=False,
        )


if __name__ == "__main__":

    example_3D()



import os
import re
import sys
import platform
import subprocess

from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion

try:
    from setuptools import setup, Extension, find_packages
except ImportError:
    print("Setuptools is required to build!")

if sys.version_info < (3, 0):
    print("Python 3.0 or higher required, please upgrade.")
    sys.exit(1)


files = ["*.so*"]


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: "
                + ", ".join(e.name for e in self.extensions)
            )

        if platform.system() == "Windows":
            cmake_version = LooseVersion(
                re.search(r"version\s*([\d.]+)", out.decode()).group(1)
            )
            if cmake_version < "3.1.0":
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = [
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + extdir,
            "-DPYTHON_EXECUTABLE=" + sys.executable,
        ]

        cfg = "Debug" if self.debug else "Release"
        build_args = ["--config", cfg]

        if platform.system() == "Windows":
            cmake_args += [
                "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), extdir)
            ]
            if sys.maxsize > 2 ** 32:
                cmake_args += ["-A", "x64"]
            build_args += ["--", "/m"]
        else:
            cmake_args += ["-DCMAKE_BUILD_TYPE=" + cfg]
            build_args += ["--", "-j2"]

        env = os.environ.copy()
        env["CXXFLAGS"] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get("CXXFLAGS", ""), self.distribution.get_version()
        )
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(
            ["cmake", ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env
        )
        subprocess.check_call(
            ["cmake", "--build", "."] + build_args, cwd=self.build_temp
        )


with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(
    name="SeismicMesh",
    version="1.2.3",
    author="Keith Roberts",
    author_email="keithrbt0@gmail.com",
    description="2D/3D serial and parallel triangular mesh generation and mesh improvement for seismology",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    setup_requires=["pybind11"],
    license_files="LICENSE",
    install_requires=[
        "numpy",
        "segyio",
        "scipy",
        "meshio",
        "h5py",
        "pybind11",
        "matplotlib",
        "mpi4py",
        "scikit-fmm",
    ],
    packages=find_packages(),
    package_data={"SeismicMesh": files},
    include_package_data=True,
    ext_modules=[
        CMakeExtension("SeismicMesh/sizing/cpp/FastHJ"),
        CMakeExtension("SeismicMesh/generation/cpp/delaunay"),
        CMakeExtension("SeismicMesh/generation/cpp/delaunay_class"),
        CMakeExtension("SeismicMesh/generation/cpp/delaunay_class3"),
        CMakeExtension("SeismicMesh/migration/cpp/cpputils"),
        CMakeExtension("SeismicMesh/geometry/cpp/fast_geometry"),
    ],
    cmdclass=dict(build_ext=CMakeBuild),
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Visualization"],
    python_requires='>=3.0',
)



import numpy as np
import copy
from mpi4py import MPI
from scipy.interpolate import RegularGridInterpolator

from .cpp import cpputils

from .. import geometry

"""
Migration routines for moving things during parallel Delaunay
"""


def localize_sizing_function(fh, h0, bbox, dim, axis, comm):
    """
    Localize the global sizing function into local chunks
    that cover the span of the local point set.
    """
    rank = comm.Get_rank()
    size = comm.Get_size()
    for r in range(0, size):
        _bbox = copy.deepcopy(bbox)
        if rank == 0:
            # form local point set
            for i in range(dim):
                if i == axis:
                    new_lims = np.linspace(_bbox[i, 0], _bbox[i, 1], size + 1)
                    _bbox[i, :] = new_lims[r : r + 2]
            grid = np.mgrid[
                tuple(slice(min - 3 * h0, max + 3 * h0, h0) for min, max in _bbox)
            ].astype(float)
            # interpolate global --> local sizing grid
            lh = fh(tuple(grid[d] for d in range(dim)))
            # updated grid vectors
            vecs = [np.arange(min - 3 * h0, max + 3 * h0, h0) for min, max in _bbox]
            # form local interpolant
            _lfh = RegularGridInterpolator(
                vecs, lh, bounds_error=False, fill_value=None
            )
            if r == 0:
                lfh = copy.deepcopy(_lfh)
                continue
            else:
                # send local interpolant to r
                comm.send(_lfh, dest=r, tag=11)
        else:
            if rank == r:
                # recv local interpolant from rank 0
                lfh = comm.recv(source=0, tag=11)
    return lfh


def localize_points(blocks, extents, comm, dim):
    """ Distribute points to local subdomains """
    rank = comm.Get_rank()
    size = comm.Get_size()
    for local in range(1, size):
        if rank == 0:
            comm.send(blocks[local], dest=local, tag=12)
        elif rank == local:
            points = np.reshape(comm.recv(source=0, tag=12), (-1, dim))
        if rank == 0:
            points = blocks[0]
        if rank != 0:
            extents = None
        extents = comm.bcast(extents, 0)
    return points, extents


def aggregate(points, faces, comm, size, rank, dim=2):
    """
    Collect global triangulation onto rank 0
    """
    # geometry.dump_mesh(points, faces, rank)

    soff_p = np.zeros((size), dtype=int)
    soff_t = np.zeros((size), dtype=int)

    soff_p[rank] = len(points)
    soff_t[rank] = len(faces)

    off_p = np.zeros((size), dtype=int)
    off_t = np.zeros((size), dtype=int)

    comm.Reduce(soff_p, off_p, op=MPI.SUM, root=0)
    comm.Reduce(soff_t, off_t, op=MPI.SUM, root=0)

    if rank == 0:
        csum_p = np.cumsum(off_p)
        gpoints = points
        gfaces = faces
    for r in np.arange(1, size):
        if rank == r:
            comm.send(points, dest=0, tag=12)
            comm.send(faces, dest=0, tag=13)
        if rank == 0:
            tmp = np.reshape(comm.recv(source=r, tag=12), (off_p[r], dim))
            tmp2 = (
                np.reshape(comm.recv(source=r, tag=13), (off_t[r], dim + 1))
                + csum_p[r - 1]
            )
            gpoints = np.append(gpoints, tmp, axis=0)
            gfaces = np.append(gfaces, tmp2, axis=0)
    if rank == 0:
        upoints, ufaces, ix = geometry.fixmesh(gpoints, gfaces, delunused=True, dim=dim)
        return upoints, ufaces
    else:
        return True, True


def enqueue(extents, points, faces, rank, size, dim=2):
    """
    Return ranks that cell sites (vertices of triangulation) need to be sent
    """
    # determine llc (lower left corner) and urc (upper right corner)
    if rank == 0:
        le = [extents[rank + 1][0:dim]]
        re = [extents[rank + 1][dim : dim * 2]]
    elif rank == size - 1:
        le = [extents[rank - 1][0:dim]]
        re = [extents[rank - 1][dim : dim * 2]]
    else:
        le = [extents[rank - 1][0:dim], extents[rank + 1][0:dim]]
        re = [extents[rank - 1][dim : dim * 2], extents[rank + 1][dim : dim * 2]]

    # add dummy box above if rank==0 or under if rank=size-1
    if rank == size - 1:
        le = np.append(le, [-999999999] * dim)
        re = np.append(re, [-999999998] * dim)

    if rank == 0:
        le = np.insert(le, 0, [-999999999] * dim)
        re = np.insert(re, 0, [-999999998] * dim)

    vtoe, ptr = geometry.vertex_to_entities(points, faces, dim=dim)

    if dim == 2:
        exports = cpputils.where_to2(points, faces, vtoe, ptr, le, re, rank)
    elif dim == 3:
        exports = cpputils.where_to3(points, faces, vtoe, ptr, le, re, rank)

    return exports


def exchange(comm, rank, size, exports, dim=2):
    """
    Exchange data via MPI using P2P comm
    """
    NSB = int(exports[0, 0])
    NSA = int(exports[0, 1])

    tmp = []
    # send points below
    if NSB != 0 and (rank != 0):  # rank 0 can't send below
        comm.send(exports[1 : NSB + 1, 0:dim], dest=rank - 1, tag=11)

    # recv  points from above
    if rank != size - 1:
        tmp = np.append(tmp, comm.recv(source=rank + 1, tag=11))

    # send points above
    if NSA != 0 and rank != (size - 1):  # topmost rank can't send to above
        # all put the topmost rank receive from above
        comm.send(exports[NSB + 1 : NSB + 1 + NSA, 0:dim], dest=rank + 1, tag=11)

    # receive points from below
    if rank != 0:
        # all but the bottommost rank receive from below
        tmp = np.append(tmp, comm.recv(source=rank - 1, tag=11))

    new_points = np.reshape(tmp, (int(len(tmp) / dim), dim))

    return new_points



from .simpplot import plot_tets, plot_facets
from .points import plot_points3
from .qual import dh_angles_hist, re_ratios_hist

__all__ = [
    "plot_tets",
    "plot_facets",
    "plot_points3",
    "dh_angles_hist",
    "re_ratios_hist",
]
